using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;

namespace WindowsFormsApplication1
{
    public class EnumUtil<T> where T : struct
    {
        #region Properties

        private static EnumUtil<T> _instance;

        private readonly Type _enumType;
        private readonly IList<KeyValuePair<T, string>> _valueDescriptionList;

        #endregion

        #region Construction

        /// <summary>
        /// Creates a new instance of <see cref="EnumUtil{T}"/>.
        /// </summary>
        private EnumUtil()
        {
            _enumType = typeof(T);

            // Validate that T is Enum
            if (!_enumType.IsEnum)
            {
                throw new InvalidOperationException(string.Format(
                    "The type '{0}' supplied to Enum<T> is not an enumeration.", _enumType.FullName));
            }

            _valueDescriptionList = Enum.GetValues(_enumType)
                .Cast<T>()
                .Select(x => new KeyValuePair<T, string>(x, ParseEnumToDescription(x)))
                .ToList();
        }

        /// <summary>
        /// Parse the enumeration value to a description. The description should be supplied with the 
        /// <see cref="DescriptionAttribute"/> on the enumeration value; or, if not existing, is the name of the enumeration 
        /// value itself.
        /// </summary>
        /// <param name="enumValue">Enumeration value to return the description of.</param>
        /// <returns>Description of th enumeration value.</returns>
        private string ParseEnumToDescription(T enumValue)
        {
            var name = Enum.GetName(_enumType, enumValue);
            var fieldInfo = _enumType.GetField(name);

            var attribute = fieldInfo.GetCustomAttributes(typeof(DescriptionAttribute), false).FirstOrDefault() as DescriptionAttribute;

            return attribute != null ? attribute.Description : name;
        }

        #endregion

        #region Properties

        /// <summary>
        /// Returns the <see cref="EnumUtil{T}"/> for the specified type.
        /// </summary>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1000:DoNotDeclareStaticMembersOnGenericTypes",
            Justification = "This is a well thought design decision.")]
        public static EnumUtil<T> Instance
        {
            get
            {
                if (_instance == null)
                {
                    _instance = new EnumUtil<T>();
                }

                return _instance;
            }
        }

        #endregion

        #region Methods

        /// <summary>
        /// Get a list of enumeration values and their corresponding user-friendly description.
        /// </summary>
        /// <returns>Enumeration values and their corresponding user-friendly description.</returns>
        public IList<KeyValuePair<T, string>> GetEnumDataSource()
        {
            return _valueDescriptionList;
        }

        /// <summary>
        /// Returns the enumeration value for a description.
        /// </summary>
        /// <param name="description">Description of enumeration value.</param>
        /// <exception cref="InvalidOperationException">No element satisfies the condition in predicate.</exception>
        /// <returns>Enumeration value.</returns>
        public T GetEnumValue(string description)
        {
            return _valueDescriptionList.First(x => x.Value == description).Key;
        }

        /// <summary>
        /// Returns the enumeration value for a description or the <paramref name="defaultValue"/> if the
        /// description doesn't exist for the enumeration.
        /// </summary>
        /// <param name="description">Description of enumeration value.</param>
        /// <param name="defaultValue">Enumeration value to return when the description doesn't exist for the enumeration.</param>
        /// <returns>Enumeration value.</returns>
        public T GetEnumValueOrDefault(string description, T defaultValue)
        {
            var enumPair = _valueDescriptionList.FirstOrDefault(x => x.Value == description);
            return enumPair.Equals(default(KeyValuePair<T, string>)) ? defaultValue : enumPair.Key;
        }

        /// <summary>
        /// Returns the description for a enumeration value.
        /// </summary>
        /// <param name="enumValue">Enumeration value.</param>
        /// <exception cref="InvalidOperationException">No element satisfies the condition in predicate.</exception>
        /// <returns>Description of enumeration value.</returns>
        public string GetEnumDescription(T enumValue)
        {
            return _valueDescriptionList.First(x => x.Key.Equals(enumValue)).Value;
        }

        #endregion
    }
}